/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2008, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

whiteLF_sound.c is for generating low passed white noise directly to
the sound card.

Sealed speaker enclosures exhibit a dual pole roll off below the
enclosure cutoff frequency, F3, (this is exactly true, in theory, for
a critically damped design, almost true for a Bessel design or
maximally flat Butterworth design, and, approximately true for a
Chebyshev design or acoustic suspension design.)

To use a sealed enclosure for measuring room acoustics, low frequency
contouring is necessary for the speaker to supply a flat white noise
response-meaning, providing the speaker with a dual pole bass boost.

This program is a modification of the white_sound(1) program, and
provides two poles and two zeros in the generated white noise
response. The pole frequency should be approximately a decade lower
than the -6 dB frequency of the speaker, (slightly below the -3 dB F3
cutoff frequency-a decade is not absolutely necessary, however,) and
extends the generated response to down -6 dB at approximately a decade
lower in frequency, i.e., it extends the flat response of the speaker
down approximately a decade, albeit with very significant power
inefficiency, (but adequate for measuring most room acoustics in a
quiet environment.)

The flat band high frequency gain and the pole frequency can be
iterated, (via stdin,) in a near field measurement of the speaker for
calibration of flat response, then measurements taken at various
locations in a room to derive the room's acoustical characteristics.

A note about the algorithm. For each time sample, a white noise
element is generated and input to two recursive pole functions in
series, (the pole frequency is specified by the double pole frequency,
p or f.) A second white noise element is generated, and the amplitude
reduced, (by the factor of the flat band high frequency gain, g,) and
summed with the output of the two recursive poles to approximate the
two zeros, thus generating a "flat band high frequency" white
noise. This technique is adequate for most purposes, and avoids the
instability issues of generating zeros in the transfer
function. Theoretically, this gives a 3 dB. error, (since the two
noise amplitudes add root-mean-square, instead of linearly,) which can
be reduced, adequately, by altering the pole frequency when doing the
near-field speaker alignment.

The commands controlling the program are read from stdin, as EOL
terminated records, and consist of a single character command followed
by a value, (if any):

    Quit: e, q
    Help: h
    Print Values: p, v
    Amplitude: a n, v n, 0 <= n <= 100, (1.0)
    Double Pole Frequency: f n, p n, 1 <= n <= 1000, (5.3)
    Flat Band High Frequency Gain: g n, 0 <= n <= 1, (0.01)

Note: the program waits for an input record on stdin before generating
the white noise, (i.e., hit the Enter key when ready.)

$Revision: 0.0$
$Date: 2008/06/29 06:01:43 $
$Id: whiteLF_sound.c,v 1.4 2008/06/29 06:01:43 john Exp john $
$Log: whiteLF_sound.c,v $
Revision 0.0  2008/02/18 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

#ifdef __STDC__

#include <float.h>

#endif

#ifndef PI /* make sure PI is defined */

#define PI 3.141592653589793 /* pi to 15 decimal places as per CRC handbook */

#endif

static char rcsid[] = "$Id: whiteLF_sound.c,v 0.0 2008/06/29 06:01:43 john Exp john $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2008, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#define TOKEN_SEPARATORS "," /* record field separators */
#define WS_SEPARATORS " \t\n\r\b,;:=" /* file record field separators */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Generate Low Passed White Noise Directly to the Sound Card\n",
    "Usage: whiteLF_sound [-d device] [-i infilename] [-v]\n",
    "    -d device, sound device, (/dev/dsp)\n",
    "    -i infilename, input filename, (stdin)\n",
    "    -v, print the program's version information\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening sound device\n",
    "Error setting sample format\n",
    "Error setting number of channels\n",
    "Error setting sample rate\n",
    "Error writing to sound device\n",
    "Error closing sound device\n",
    "Error overflow buffer\n",
    "Error getting status of input file\n",
    "Error selecting input file\n"
    "Error opening input file\n",
    "Error closing input file\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define EOPEN 2
#define EIOF 3
#define EION 4
#define EIOS 5
#define EWRITE 6
#define ECLOSE 7
#define EOVRFL 8
#define EISTAT 9
#define EISEL 10
#define EIOPEN 11
#define EICLOSE 12

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static int strtoken (char *string, char *parse_array, char **parse, const char *delim); /* parse a record based on sequential delimiters */
static int white_gen (char *device, int infile); /* generate a white noise through the sound card */
static int interpret (int fields, char *token[]); /* interpret a command from stdin */

#else

static void print_message (); /* print any error messages */
static int strtoken ();  /* parse a record based on sequential delimiters */
static int white_gen (); /* generate a white noise through the sound card */
static int interpret (); /* interpret a command from stdin */

#endif

static int sample_rate = 44100; /* sound sample rate */

static double k1, /* coefficient k1 in the recursive filter */
              k2, /* coefficient k2 in the recursive filter */
              p = (double) 5.3, /* pole frequency = 5.3 Hz. */
              g = (double) 0.01, /* flat band high frequency gain */
              amplitude = (double) 1.0; /* amplitude to generate, defaults to -15 dBFS */

char sound_device[] = "/dev/dsp"; /* sound device */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char *device = sound_device, /* sound device */
         inchar, /* character from input file */
         inbuffer[BUFLEN + 1], /* input read buffer */
         parsebuffer[BUFLEN], /* parsed i/o buffer */
         *in_name = (char *) 0, /* input file name */
         *token[BUFLEN / 2], /* reference to tokens in parsed i/o buffer */
         *ref, /* character reference in input buffer */
         *sentinel = &inbuffer[BUFLEN]; /* reference to sentinel character in input buffer */

    int retval = NOERROR, /* return value, assume no error */
        fields, /* number of fields */
        infile = STDIN_FILENO, /* input file descriptor */
        c; /* command line switch */

    ssize_t cnt_in = 0; /* count of characters read */

    *sentinel = '\0'; /* sentinel character in input buffer */

    p = p / (double) sample_rate; /* pole frequency = 5.3 Hz. */
    k1 = exp (- ((double) 2.0 * (double) PI * p)); /* calculate the coefficient k1 in the recursive filter */
    k2 = (double) 1.0 - k1; /* calculate the coefficient k2 in the recursive filter */

    while ((retval == NOERROR) && ((c = getopt (argc, argv, "d:i:v")) != EOF)) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'd': /* request for sound device? */

                device = optarg; /* yes, save the sound device */
                break;

            case 'i': /* request for input file name? */

                in_name = optarg; /* yes, save the input file name */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = EIOPEN; /* assume error opening file */

        if ((infile = (in_name == (char *) 0) ? infile : open (in_name, O_RDWR)) >= 0) /* open the input file */
        {
            retval = NOERROR; /* assume no error */
            ref = inbuffer; /* reset the character reference in input buffer */
            *ref = '\0'; /* terminate the record */

            while ((retval == NOERROR) && ((cnt_in = read (infile, &inchar, sizeof (char))) == sizeof (char))) /* read characters from the input file */
            {
                retval = EOVRFL; /* assume error overflow buffer */

                if (ref < sentinel) /* buffer overflow? */
                {
                    *ref = inchar; /* save the character in the input buffer */
                    ref ++; /* prepare for next character reference in input buffer */
                    *ref = '\0'; /* terminate the record; may overwrite the sentinel '\0' */
                    retval = NOERROR; /* assume no error */

                    if (inchar == '\n') /* character an EOL? */
                    {

                        if ((fields = strtoken (inbuffer, parsebuffer, token, WS_SEPARATORS)) != 0) /* parse the record into fields, skip null records */
                        {

                            if (interpret (fields, token) != 0) /* interpret the command from stdin */
                            {
                                break; /* quit command issued */
                            }

                        }

                        ref = inbuffer; /* reset the character reference in input buffer */
                        *ref = '\0'; /* terminate the record */
                        retval = white_gen (device, infile); /* generate white noise through the sound card */
                    }

                }

            }

            if (infile != STDIN_FILENO) /* input file open? */
            {

                if (close (infile) == 0) /* close the input file */
                {
                    infile = STDIN_FILENO; /* reset the input file descriptor */
                }

            }

        }

        retval = (cnt_in == 0) ? NOERROR : retval; /* a count of zero, and error reading input file was an EOF or ^D, a normal exit */
    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

Parse a record based on sequential delimiters.

int strtoken (char *string, char *parse_array, char **parse, const char *delim);

I) Parse a character array, string, into an array, parse_array, using
consecutive characters from delim as field delimiters, point the
character pointers, token, to the beginning of each field.

Returns the number of fields parsed.

*/

#ifdef __STDC__

static int strtoken (char *string, char *parse_array, char **parse, const char *delim)

#else

static int strtoken (string, parse_array, parse, delim)
char *string;
char *parse_array;
char **parse;
char *delim;

#endif

{
    int tokens = 0;

    (void) strcpy (parse_array, string); /* copy the string */

    parse[tokens] = strtok (parse_array, delim); /* get the 1st field */

    while (parse[tokens] != 0) /* get the remaining fields */
    {
        parse[++ tokens] = strtok ((char *) 0, delim);
    }

    return (tokens); /* return the number of tokens parsed */
}

/*

Generate white noise through the sound card.

static int white_gen (char *device, int infile);

I) Data structure:

    1) device is the sound device to use for the white noise.

    2) infile is the input file to read white noise amplitude.

Returns any error, NOERROR on success.

*/

#ifdef __STDC__

static int white_gen (char *device, int infile)

#else

static int white_gen (device, infile)
char *device;
int infile;

#endif

{
    short val; /* value of white noise signal sample, (always the floor (sample_value)) */

    int retval = NOERROR, /* return value, assume no error */
        tmp, /* temporary variable */
        pending_ret, /* pending amplitude flag, 0 = no, 1 = yes, -1 = error */
        cnt, /* sample counter */
        fd = -1; /* sound device file descriptor */

    double sample_value = (double) 0.0, /* the value of white noise signal sample */
           flatband, /* flat band white noise */
           lastvalue1 = (double) 0.0, /* last value of first pole */
           lastvalue2 = (double) 0.0; /* last value of second pole */

    struct timeval timeout; /* time value structure for select(2) determination of pending amplitude */

    fd_set fds; /* select () struct for determination of pending amplitude */

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = EOPEN; /* assume error opening sound device */

        if ((fd = open (device, O_WRONLY, (mode_t) 0)) != -1) /* open the sound device */
        {

            /* set sample format to 16 bits */

            tmp = AFMT_S16_NE; /* native 16 bits */
            retval = EIOF; /* assume error setting sample format */

            if (ioctl (fd, SNDCTL_DSP_SETFMT, &tmp) != -1) /* native 16 bits */
            {

                if (tmp == AFMT_S16_NE) /* 16 bits supported? */
                {

                    /* set number of channels */

                    tmp = 1;
                    retval = EION; /* assume error setting number of channels */

                    if (ioctl (fd, SNDCTL_DSP_CHANNELS, &tmp) != -1) /* mono */
                    {

                        if (tmp == 1) /* mono supported? */
                        {

                            /* set the sample rate */

                            retval = EIOS; /* assume error setting sample rate */

                            if (ioctl (fd, SNDCTL_DSP_SPEED, &sample_rate) != -1)
                            {
                                retval = NOERROR; /* assume no error */
                                cnt = sample_rate; /* set the sample counter for one second */

                                while (retval == NOERROR) /* for each time sample */
                                {
                                    sample_value = (((double) rand () / (double) RAND_MAX) - (double) 0.5) * (double) 65534; /* generate a random number between -32767 and +32767 */

                                    /* two poles in series: */

                                    lastvalue1 = sample_value * k2 + lastvalue1 * k1; /* compute the last value of the first pole filtered white noise */
                                    lastvalue2 = lastvalue1 * k2 + lastvalue2 * k1; /* compute the last value of the second pole filtered white noise */

                                    /* two zeros make the flatband: */

                                    flatband = (((double) (rand () / (double) RAND_MAX) - (double) 0.5) * (double) 65534.0) * g; /* generate the flat band high frequency noise */
                                    sample_value = (lastvalue2 + flatband) * amplitude; /* add the flat band noise */

                                    if ((sample_value < (double) SHRT_MIN) || (sample_value > (double) SHRT_MAX)) /* test for short integer over/under flow */
                                    {
                                        (void) fprintf (stderr, "(Overflow)\n"); /* print the error if it is */
                                    }

                                    val = (short) floor (sample_value); /* value of white noise signal sample, (always the floor (sample_value)) */
                                    retval = EWRITE; /* error writing to sound device */

                                    if (write (fd, &val, sizeof (short)) == (ssize_t) sizeof (short)) /* write the record to the sound device */
                                    {
                                        retval = NOERROR; /* assume no error */

                                        if (cnt -- < 0) /* one second passed? */
                                        {
                                            cnt = sample_rate; /* set the sample counter for one second */
                                            timeout.tv_sec = (long) 0; /* time value structure for select(2) determination of pending amplitude, seconds */
                                            timeout.tv_usec = (long) 1; /* time value structure for select(2) determination of pending amplitude, microseconds; place holder, only-most system will not time to one microsecond */
                                            FD_ZERO (&fds); /* null the select () struct for select(2) determination of pending amplitude */
                                            FD_SET (infile, &fds); /* install the select () struct for select(2) determination of pending amplitude */

                                            if ((pending_ret = select (infile + 1, &fds, (fd_set *) 0, (fd_set *) 0, &timeout)) != 0) /* check for pending data on infile, i.e., determine if pending amplitude */
                                            {
                                                retval = EISEL; /* assume error selecting input file */

                                                if (pending_ret == 1) /* pending amplitude? */
                                                {
                                                    retval = NOERROR; /* assume no error */
                                                    break; /* interrupt the current white noise, and do the next amplitude */
                                                }

                                            }

                                        }

                                    }

                                }

                            }

                        }

                    }

                }

            }

            if (close (fd) == -1) /* close the sound device */
            {

                if (retval == NOERROR) /* any previous errors? */
                {
                    retval = ECLOSE; /* assume error closing sound device */
                }

            }

        }

    }

    return (retval); /* return any errors */
}

/*

Interpret a command from stdin.

static int interpret (int fields, char *token[]);

I) Data structure:

    1) fields is the number of fields in the command.

    2) token is a reference to the tokens in the command.

Returns NOERROR on success, 1 if the command was to quit.

*/

#ifdef __STDC__

static int interpret (int fields, char *token[])

#else

static int interpret (int fields, char *token[])
int fields;
char *token[];

#endif

{
    int retval = NOERROR, /* return value, assume no error */
        print_help = 1; /* print help? NOERROR = no, anything else = yes, assume yes */

    double tmp; /* temporary double storage */

    switch (fields) /* either 1 or 2 fields */
    {

        case 1: /* one field command */

            if (token[0][1] == '\0') /* single letter command? */
            {

                switch (token[0][0]) /* which command? */
                {

                    case 'e': /* request to quit? */
                    case 'E':
                    case 'q':
                    case 'Q':
                    case 'x':
                    case 'X':

                        retval = 1;
                        print_help = NOERROR; /* print help? NOERROR = no, anything else = yes */
                        break;

                    case 'h': /* request for help? */
                    case 'H':

                        break;

                    case 'p': /* request for print values? */
                    case 'P':
                    case 'v':
                    case 'V':

                        (void) fprintf (stderr, "Amplitude: %f\n", amplitude);
                        (void) fprintf (stderr, "Double Pole Frequency: %f\n", p * (double) sample_rate);
                        (void) fprintf (stderr, "Flat Band High Frequency Gain: %f\n", g);
                        print_help = NOERROR; /* print help? NOERROR = no, anything else = yes */
                        break;

                    default: /* none of the above, print the error */

                        break;
                }

            }

            break;

        case 2: /* two field command */

            if (token[0][1] == '\0') /* single letter command? */
            {

                switch (token[0][0]) /* which command? */
                {

                    case 'a': /* request for amplitude? */
                    case 'A':
                    case 'V':
                    case 'v':

                        tmp = atof (token[1]); /* yes, save the amplitude */

                        if ((tmp >= (double) 0.0) && tmp <= (double) 100.0) /* sanity check */
                        {
                            amplitude = tmp; /* save the amplitude */
                            print_help = NOERROR; /* print help? NOERROR = no, anything else = yes */
                        }

                        break;

                    case 'f': /* request for double pole frequency? */
                    case 'F':
                    case 'p':
                    case 'P':

                        tmp = atof (token[1]); /* yes, save the double pole frequency */

                        if ((tmp >= (double) 1.0) && (tmp <= (double) 1000.0)) /* sanity check */
                        {
                            p = tmp / (double) sample_rate; /* pole frequency = 5.3 Hz. */
                            k1 = exp (- ((double) 2.0 * (double) PI * p)); /* calculate the coefficient k1 in the recursive filter */
                            k2 = (double) 1.0 - k1; /* calculate the coefficient k2 in the recursive filter */
                            print_help = NOERROR; /* print help? NOERROR = no, anything else = yes */
                        }

                        break;

                    case 'g': /* request for flat band high frequency gain? */
                    case 'G':

                        tmp = atof (token[1]); /* yes, save the flat band high frequency gain */

                        if ((tmp >= (double) 0.0) && (tmp <= (double) 1.0)) /* sanity check */
                        {
                            g = tmp; /* save the flat band high frequency gain */
                            print_help = NOERROR; /* print help? NOERROR = no, anything else = yes */
                        }

                        break;

                    default: /* none of the above, print the error */

                        break;

                }

            }

            break;

        default: /* none of the above, print the error */

            break;

    }

    if (print_help != NOERROR) /* print help? NOERROR = no, anything else = yes */
    {
        (void) fprintf (stderr, "Quit: e, q, x\n");
        (void) fprintf (stderr, "Help: h\n");
        (void) fprintf (stderr, "Print Values: p, v\n");
        (void) fprintf (stderr, "Amplitude: a n, v n, 0 <= n <= 100, (1.0)\n");
        (void) fprintf (stderr, "Double Pole Frequency: f n, p n, 1 <= n <= 1000, (5.3)\n");
        (void) fprintf (stderr, "Flat Band High Frequency Gain: g n, 0 <= n <= 1, (0.01)\n");
    }

    return (retval); /* return any errors */
}
